function [Y,M,de] = desquelch(X,SR,T,HPF,VERBOSE,standofftime,minchunktime)
% [Y,M] = desquelch(X,SR,T,HPF,VERBOSE,standofftime,minchunktime)
%    Find "squelched" regions in waveform X (regions of very low
%    amplitude).  Remove them and close up to return Y; M is a map
%    of which intervals were removed (rows of [start end] time). 
%    T is threshold for declaring "squelch" in dB (absolute).
%    Default is -70.0.
%    VERBOSE reports actual median energy in the region that is squelched
% 2012-05-15 Dan Ellis dpwe@ee.columbia.edu

if nargin < 3;  T = -70.0; end
if nargin < 4;  HPF = 0; end
if nargin < 5;  VERBOSE = 0; end
if nargin < 6;  standofftime = 0.010; end
if nargin < 7;  minchunktime = 0.010; end

%disp(['desquelch=',num2str(T)]);

% Calculate energy over chunks this long
smoothtime = 0.002;

hchunksamps = round(smoothtime/2 * SR);
chunksamps = 2*hchunksamps + 1; % odd

cwin = ones(chunksamps,1)/chunksamps;
kpx = hchunksamps + [1:length(X)];

if HPF == 0
  % no filtering
  de = conv(cwin,X.^2);
  de = de(kpx);
else
  % IIR highpass forward and backward, take min env
  [b,a] = butter(6,HPF/(SR/2), 'high');  
  def = conv(cwin,filter(b,a,X).^2);
  deb = flipud(conv(cwin,filter(b,a,flipud(X)).^2));
  de = min(def(kpx),deb(kpx));
end

% to dB
de = 10*log10(de);


% Find stretches more than T below max
%maxd = max(de);
%dmask = de > (maxd+T);
% Have it as an absolute threshold, to handle pure-silence files
dmask = de > T;

% Build a set of transitions
tx10 = find( ([1;dmask]==1) & ([dmask;1]==0) )-1;
tx01 = find( ([1;dmask]==0) & ([dmask;1]==1) );
M = [tx10,tx01]/SR;

% remove regions that are too short
%minchunktime = 0.010;
% .. and grow deleted regions this far at each end
%standofftime = 0.010;
% now an input arg

longenough = find( (M(:,2)-M(:,1)) >= minchunktime);

M = M(longenough,:);

% Skip this if M is empty (no excluded regions)
if size(M,1) == 0

  Y = X;

else

  % apply standoff
  M(:,1) = max(0, M(:,1)-standofftime);
  M(:,2) = min(length(de)/SR, M(:,2)+standofftime);

  dmask = find(dmask);

  % reconstruct index
  dmask2 = zeros(length(X),1);
  dmask2(1:round(M(:,1)*SR)) = 1;
  for i = 1:(size(M,1)-1)
    dmask2((round(SR*M(i,2))+1):round(SR*M(i+1,1))) = 1;
  end
  dmask2((round(SR*M(end,2))+1):length(X)) = 1;

  % reconstruct M
  tx10 = find( ([1;dmask2]==1) & ([dmask2;1]==0) )-1;
  tx01 = find( ([1;dmask2]==0) & ([dmask2;1]==1) );
  M = [tx10,tx01]/SR;

  if VERBOSE
    % actual energy in squelched regions
    mediansquelchE = median(de(find(dmask2==0)));
    disp(['desquelch: median energy of squelched=', ...
          num2str(mediansquelchE),' dB']);
  end
  
  dmask2 = find(dmask2);

  % Collapse the signal
  Y = X(dmask2);
  
end
