function findNTs(varargin)
% findNTs [varargin] audiofile1 ...
%   apply automatic NT detection to files by batch
%   -labelpath ''         Path to label files
%   -labelext '.txt'      Extension of label files
%   -labeltype 'ldc'      Label file format ('ldc'/'3col'/'segmap')
%   -disp 0/1             Optional graphic display (0)
%   -NTpath ''            Save labels from findNT to this directory
%   -NText '.fnt'         extension for saved NT label files
%   -NTpatch 0/1          Make 1 to have output NTs include S, NS from input
%   -NTlowE 0/1           If 1, look for low-energy NT cluster (not high)
%   -tries 5              Repeat findNT this many times to find best
%   -block 10             Max number of files to process at once
% 2012-02-22 Dan Ellis dpwe@ee.columbia.edu

VERSION = 0.35;
DATE = 20121004;

% Parse out the optional arguments
[FILELIST, NFILES, NSKIP, AUDIOPATH, AUDIOEXT, ...
 LABELPATH, LABELEXT, LABELSFORMODEL, ...
 INLABTYPE, OUTLABTYPE, ...
 SUMMARY, ...
 NTPATH, NTEXT, ...
 NTPATCH, NTLOWE, ...
 NTMINPROP, NTVOLTHRESH, ...
 NTCOLLAR, ...
 USEVX, VXTHRESH, ...
 MINKLSEP, MINMODALSCORE, ...
 DESQUELCH, ONLYDESQUELCH, ...
 TRIES, ...
 BLOCK, ...
 NCOMP, NFRAME, ...
 CLSTFACT, CLSTPCNTILE, ...
 MERGE_KLTHRESH, VAR_SCALEUP, TXPROB, ...
 MODELSIN, MODELSOUT, ...
 VIEWMODELS, VIEWNTS, ...
 POSTOUT, ...
 LOG, ...
 RESETRAND, ...
 SLOWDECODE, ...
 VERBOSE, SILENT, HELP, ...
 XTRA] = ...
    process_options(varargin, ...
                    '-filelist', '', ...
                    '-nfiles', 0, ...
                    '-nskip', 0, ...
                    '-audiopath', '', ...
                    '-audioext', '.flac', ...
                    '-labelpath', '', ...
                    '-labelext', '.txt', ...
                    '-labelsformodel', 0, ...
                    '-inlabtype', 'ldc', ...
                    '-outlabtype', '3col', ...
                    '-summary', '', ...
                    '-NTpath', '.', ...
                    '-NText', '.fnt', ...
                    '-NTpatch', 0, ...
                    '-NTlowE', 0, ...
                    '-NTminprop', 0.0075, ...
                    '-NTvolthresh', 2.0, ...
                    '-NTcollar', 0.020, ...
                    '-usevx', 1, ...
                    '-vxthresh', Inf, ...
                    '-minklsep', 0, ...
                    '-minmodalscore', 6.0, ...
                    '-desquelch', -75, ...
                    '-onlydesquelch', 0, ...
                    '-tries', 5, ...
                    '-block', 10, ...
                    '-ncomp', 16, ...
                    '-nframe', 10000, ...
                    '-clstfact', 1.5, ...
                    '-clstpcntile', 0.1, ...
                    '-merge_klthresh', 10, ...
                    '-var_scaleup', 0, ...
                    '-txprob', 0, ...
                    '-modelsin', '', ...
                    '-modelsout', '', ...
                    '-viewmodels', 0, ...
                    '-viewnts', 0, ...
                    '-postout', '', ...
                    '-log', '', ...
                    '-resetrand', 0, ...
                    '-slowdecode', 0, ...
                    '-verbose', 0, '-silent', 0, '-help', 0);

%ddisp(['********** findNTs v',num2str(VERSION),' of ',num2str(DATE), ...
%       ' **********'], VERBOSE);

if HELP
  disp(['findNTs v',num2str(VERSION),' of ',num2str(DATE)]);
  disp('usage: findNTs [varargin] audiofile1 ...');
  disp('  Batch automatic identification of NT regions.');
  disp('  -filelist <name>      File containing additional audio file names');
  disp('  -nfiles 0             Process only this many files (if nonzero)');
  disp('  -nskip 0              Skip this many files from start of list');
  disp('  -audiopath <dir>      Directory containing audio files');
  disp('  -audioext ''.flac''     Extension of audio files');
  disp('  -labelpath <dir>      Directory containing label files');
  disp('  -labelext ''.txt''      Extension of label files');
  disp('  -labelsformodel 0     Use labels to help choose the model');
  disp('  -inlabtype ''ldc''      Type of input label files (ldc/segmap/3col)');
  disp('  -outlabtype ''3col''    Type of output label files (ldc/segmap/3col)');
  disp('  -summary <fname>      Write summary results to this file');
  disp('  -NTpath <dir>         Directory to write out findNT labels');
  disp('  -NText ''.fnt''         Extension for findNT label files');
  disp('  -NTpatch 0            0=write just NT labs; 1=include orig S/NS');
  disp('  -NTlowE 0             if 1 choose low energy cluster as NT');
  disp('  -NTcollar 0.020       Grow NT regions this much at each end');
  disp('  -NTminprop 0.0075     Min proportion of frames for NT cluster');
  disp('  -NTvolthresh 2.0      Factor larger than smallest for NT volume');
  disp('  -usevx 1              Calculate and use the (slow) voicing feature');
  disp('  -vxthresh Inf         Oblige NT to have voicing below this value');
  disp('  -minklsep 0           Dont use NT model if too similar to rest');
  disp('  -minmodalscore 0      Dont use NT model if mode in data too weak');
  disp('  -desquelch -75        Strip out silent blocks below this many dB');
  disp('  -onlydesquelch 0      Only use desquelch (not GMMs) to find NTs');
  disp('  -tries 5              Chose found NTs as best of this many runs');
  disp('  -block 10             Process (up to) this many files in a block');
  disp('  -ncomp 16             How many components to use in GMMs');
  disp('  -nframe 10000         How many frames to sample per model');
  disp('  -clstfact 1.5         Factor over decile distance for similar');
  disp('  -clstpcntile 0.1      Percentile as similar distance reference');
  disp('  -merge_klthresh 10    Merge gaussians whose KL dist is below this');
  disp('  -var_scaleup 0        "Guard" gaussian''s variance scaleup (0=none)');
  disp('  -txprob 0             Transition probability for decode (0=use default)');
  disp('  -modelsin <fname>     Read set of models from this file');
  disp('  -modelsout <fname>    Write models created to this file');
  disp('  -viewmodels 0         Inspect models one by one?');
  disp('  -viewnts 0            Display final NT regions on spectrogram');
  disp('  -postout <fstem>      Write actual NT posteriors to <fstem>-<file>.mat');
  disp('  -log <logfile>        Write log to this file');
  disp('  -resetrand 0          Reset random number stream?');
  disp('  -slowdecode 0         Disable MEX in viterbi decode?');
  disp('  -verbose 0            More progress messages');
  disp('  -silent 0             No messages at all');
  return
end

% Maybe reset random stream to known state
if RESETRAND > 0
  resetrand();
end

% convert label type names to index
INLABCODE = labtype2code(INLABTYPE);
OUTLABCODE = labtype2code(OUTLABTYPE);

if length(SUMMARY) > 0
  sumfp = fopen(SUMMARY, 'w');
else
  sumfp = 1; % stdout
end

if length(LOG) > 0
  logfp = fopen(LOG, 'w');
  if VERBOSE
    VERBOSE = logfp;
  else
    VERBOSE = -logfp;
  end
else
  logfp = 0;
end

%ddisp(['********** findNTs v',num2str(VERSION),' of ',num2str(DATE), ...
%       ' **********'], VERBOSE);
%ddisp(['Arguments: '], VERBOSE);
%ddisp([join_cells(varargin,' ')], VERBOSE);
if sumfp
  fprintf(sumfp, '%s\n', ['********** findNTs v',num2str(VERSION),' of ',...
                  num2str(DATE),' **********']);
  fprintf(sumfp, 'Arguments:\n');
  fprintf(sumfp, '%s\n', [join_cells(varargin,' ')]);
end

FILES = XTRA;

if length(FILELIST)
  if exist(FILELIST, 'file')
    FILES = [listfileread(FILELIST), FILES];
  else
    error(['Input file list file ',FILELIST,' not found']);
  end
end

nfiles = length(FILES);
if nfiles == 0
  error('No input files specified');
end

% optional drop of initial files (for working with lists)
if NSKIP > 0
  FILES = FILES((NSKIP+1):end);
  nfiles = nfiles - NSKIP;
end

% optional limit on the number of files to process
% (to subset a list)
if NFILES > 0 && nfiles > NFILES
  FILES = FILES(1:NFILES);
  nfiles = NFILES;
end


% Always pause when displaying for multiple files
if nfiles > 1
  PAUSE = 1;
else
  PAUSE = 0;
end

% Backwards compatibility: strip wavfile extension from end of
% input file names
lenext = length(AUDIOEXT);
if lenext > 0
  for i = 1:nfiles
    if strcmp(FILES{i}(end-(lenext-1):end),AUDIOEXT) == 1
      FILES{i} = FILES{i}(1:end-lenext);
    end
  end
end

% Make sure audiopath and labelpath end in slash if not empty
if length(AUDIOPATH) > 0
  AUDIOPATH = ensuresuffix(AUDIOPATH, '/');
end
if length(LABELPATH) > 0
  LABELPATH = ensuresuffix(LABELPATH, '/');
end
if length(NTPATH) > 0
  NTPATH = ensuresuffix(NTPATH, '/');
end


targetsr = 8000;
frametime = 0.010; % we assume
% Define global label ordering
IX = {'S','NS','NT','RX','RM'};
SPx = 1; NSx = 2; NTx = 3; RXx = 4; RMx = 5;

% Algorithm: we gather up files into blocks of BLOCK most similar
% files, build models on those, then decode every file using the
% most similar model

% Calculate mutual similarity of files by building gaussian models
% for each of them

if ONLYDESQUELCH

  for i = 1:nfiles
    data = NT_get_data(addprefixsuffix(FILES(i),AUDIOPATH,AUDIOEXT), ...
                       NFRAME, USEVX, DESQUELCH, VERBOSE);
  end
  modelfor = zeros(nfiles,1);
  modeldist = zeros(nfiles,1);
  
else
  

  % Build GMM models for each input track
  goodmodels = zeros(nfiles,1);
  for i = 1:nfiles
    data = NT_get_data(addprefixsuffix(FILES(i),AUDIOPATH,AUDIOEXT), ...
                       NFRAME, USEVX, DESQUELCH, VERBOSE);
    %disp(['data has ',num2str(size(data,1)),' rows']);
    if size(data,1) < 100
      ddisp(['Too few frames to process in ', FILES{i}],VERBOSE)
      filemodels(i) = struct('mean',[],'cov',[],'weight',[],'diag',0);
    else
      filemodels(i) = gmm_est(data, NCOMP);
      goodmodels(i) = 1;
    end
  end
  goodmodels = find(goodmodels);

  if length(MODELSIN) == 0
    % We need to create new models from the clustered input data

    % full-order distance matrix
    gd = gmm_distance(filemodels(goodmodels));
    % fix degenerate NaN models
    gd(find(eye(length(goodmodels)))) = 0;

    % Routine to build the set of models from clustered files
    if LABELSFORMODEL
      BUILDLABELPATH = LABELPATH;
      BUILDLABELEXT = LABELEXT;
    else
      BUILDLABELPATH = '';
      BUILDLABELEXT = '';
    end
    
    models = NT_build_all_models(gd, ...
                                 FILES(goodmodels), ...
                                 AUDIOPATH, AUDIOEXT, CLSTFACT, ...
                                 CLSTPCNTILE, BLOCK,...
                                 NFRAME, TRIES, ...
                                 NTLOWE, NCOMP, ...
                                 MERGE_KLTHRESH, VAR_SCALEUP, ...
                                 BUILDLABELPATH, BUILDLABELEXT, INLABCODE,...
                                 TXPROB, USEVX, DESQUELCH, ...
                                 NTMINPROP, NTVOLTHRESH, ...
                                 VERBOSE, VIEWMODELS);

    % Done building models; maybe save them
    if length(MODELSOUT) > 0
      format = '20120430';
      save(MODELSOUT,'models','format');
      ddisp(['Models saved to ',MODELSOUT], SILENT==0); % instead of VERBOSE 
    end

  else
    % MODELSIN specified - Load pre-defined models
    M = load(MODELSIN);
    models = M.models;

  end

  % Figure best model for each file
  modeldist = zeros(nfiles);
  modelfor = zeros(nfiles);
  gd = gmm_distance(models, filemodels(goodmodels));
  [modeldist(goodmodels), modelfor(goodmodels)] = min(gd,[],1);

end
  
  
% Now decode each file with associated model
for i = 1:nfiles

  % Load the features (again?)
  [data,mlen,desm] = NT_get_data(addprefixsuffix(FILES(i),AUDIOPATH,AUDIOEXT), ...
                                 0,USEVX,DESQUELCH,VERBOSE);

  %squelchprop = max(0,sum(diff(desm))/(size(data,1)*frametime));
  squelchtime = max(0,sum(diff(desm,1,2)));
  
  if ONLYDESQUELCH
    nts = zeros(0,3);
  else    
    if modelfor(i) == 0
      ddisp(['**** No model - decode skipped'], VERBOSE);
      nts = zeros(0,3);
      model = struct();
      model.name = '<none>';
      vxval = 0;
      modalscore = 0;
    else
    
      model = models(modelfor(i));

      vxval = model.mean(2,model.NT);
      % modal score for this file alone?
      %modalscore = gmm_modal_score(model, data, model.NT);
      % .. or from entire cohort in model building?
      modalscore = model.modalscore;
    
      if model.KLsep < MINKLSEP
        % not good enough!
        ddisp(['**** Model KLsep ',num2str(model.KLsep),' less than ', ...
               num2str(MINKLSEP), ' - decode skipped'], VERBOSE);
        nts = zeros(0,3);
      elseif modalscore < MINMODALSCORE
        ddisp(['**** Model modalscore ',num2str(modalscore),' less than ', ...
               num2str(MINMODALSCORE), ' - decode skipped'], VERBOSE);
        nts = zeros(0,3);
      elseif USEVX && (vxval > VXTHRESH)
        ddisp(['**** Model VX value ',num2str(vxval),' more than ', ...
               num2str(VXTHRESH), ' - decode skipped'], VERBOSE);
        nts = zeros(0,3);
      else
        % Decode them with the matching model
        [nts,ntpost] = NT_decode(data, model, model.NT, TXPROB, SLOWDECODE);
        if length(POSTOUT) 
          [Fpth,Fstem,Fext] = fileparts(FILES{i});
          postoutfile = [POSTOUT,'-',Fstem,'.mat'];
          save(postoutfile, 'ntpost');
          disp(['Posteriors saved to ',postoutfile]);
        end
      end
    end
      
    ddisp(['File ',num2str(i), ' ', FILES{i},' best model ', ...
           num2str(modelfor(i)), ' ', model.name, ...
          ' dist=',num2str(modeldist(i)), ...
          ' modalscore=',num2str(modalscore), ...
          ' vxval=',num2str(vxval)], VERBOSE);

    % Inspect?
    if modelfor(i) > 0 && VIEWMODELS
      gmm_plotdata(data, model, model.NT, PAUSE);
    end
  
  end

  % reinstate desquelch gaps
  nts = NT_resquelch(nts,desm);
  % apply collar (even to desquelch)
  nts = labels_collarize(nts, NTCOLLAR);
  
  if length(nts) == 0
    ddisp(['Warn: zero NTs found for ', FILES{i}], VERBOSE);
  end
  
  if sumfp
    fprintf(sumfp, ...
            '%s: SQ %5.1f LT %5.1f BM %2d MDST %5.3f FNTT %5.1f', ...
            FILES{i}, squelchtime, frametime*size(data,1), ...
            modelfor(i), modeldist(i), ...
            diff(sum(nts(:,[1 2]),1)));
  end
  
  % Check against existing labels?
  if length(LABELPATH)>0
    LABFILE = addprefixsuffix(FILES{i},LABELPATH,LABELEXT);
    if exist(LABFILE,'file') > 0
      [L,IX,XTRA] = read_label_file(LABFILE, IX, INLABCODE, 'NT');
      % Compare found NTs against NTs/RXs/RMs from labelfile
      NRIX = find(L(:,3)==NTx | L(:,3)==RXx | L(:,3)==RMx);
      NRL = L(NRIX,:);
      NRL(:,3) = 1;  % dummy
      [SNT,CNT] = compare_labels(nts, NRL);
      % Also compare against SP labels from file
      SPIX = find(L(:,3) == SPx);
      SPL = L(SPIX,:);
      [SSP,CSP] = compare_labels(nts, SPL);
      % Report overlaps to summary file
      if sumfp
        fprintf(sumfp, ...
                ' TT %5.1f NTT %5.1f FNTNT %7.3f FNTNT%% %5.2f FNTSP %7.3f FNTSP%% %5.2f', ...
                max(L(:,2)), diff(sum(NRL(:,[1 2]),1)), ...
                CNT(2,2), SNT*100, CSP(2,2), SSP*100);
      end
      if VIEWNTS
        % Plot actual labels over spectrogram
        plot_label_comparison(addprefixsuffix(FILES{i},AUDIOPATH,AUDIOEXT), ...
                              targetsr,NRL,nts,PAUSE);
      end
    else
      disp(['warn: label file ',LABFILE,' not found']);
      L = [];
    end
  else
    L = [];
    if VIEWNTS
      % Plot actual labels over spectrogram
      plot_label_comparison(addprefixsuffix(FILES{i},AUDIOPATH,AUDIOEXT), ...
                            targetsr,[],nts,PAUSE);
    end
  end
  if sumfp; fprintf(sumfp, '\n'); end
    
  % Maybe write out new NT label file
  if length(NTPATH) > 0
    NTFNAME = addprefixsuffix(FILES{i}, NTPATH, NTEXT);
    % make new labels be actual NTs
    nts(:,3) = NTx;
    % Are we going to insert these over the top of the existing labels?
    if NTPATCH && length(L) > 0
      % grab some XTRA first, in case everything gets deleted
      newx = XTRA(1);
      % delete NT and RX from original labels
      [L1,XTRA] = delete_labels(L,NTx,XTRA);
      [L2,XTRA] = delete_labels(L1,RXx,XTRA);
      [Ln,XTRA] = delete_labels(L2,RMx,XTRA);
      % merge new NTs with original S and SP
      % setup default 11col info indicating where label came from
      if length(XTRA) > 1
        % prefer XTRA from an actual non-NT segment
        newx = XTRA(1);
      end
      newx{1}{6} = 'findNTs';
      [outnts,XTRO] = merge_labels(Ln,nts,XTRA,newx);
    else
      outnts = nts;  XTRO = [];
    end
    % write out
    mymkdir(fileparts(NTFNAME));
    write_label_file(outnts, NTFNAME, IX, OUTLABCODE, XTRO);
    ddisp([num2str(length(outnts)),' labels saved to ',NTFNAME], SILENT==0);
  end

end

if sumfp > 1
  fclose(sumfp);
end
if logfp > 1
  fclose(logfp);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function ddisp(msg, verb)
if verb > 0
  disp(msg);
end
% verb also carries file descriptor of log file
if verb ~= 0
  if verb ~= 1 && verb ~= 2
    fprintf(abs(verb), '%s\n', msg);
  end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function s = join_cells(c,j)
% flatten a cell array of strings into a single string
if nargin < 2; j = ''; end
s = c{1};
for i = 2:length(c);
  s = [s,j,c{i}];
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function plot_label_comparison(FILE,targetsr,NRL,nts,PAUSE)
% plot the NT label comparison
%subplot(414)
specgram(audioread(FILE,targetsr),round(0.02*targetsr),targetsr);
colormap(1-gray);
[p,cleanstem,e] = fileparts(FILE);
title(cleanstem,'interpreter','none');
caxis(max(caxis)+[-80 0]);
colorbar
hold on; 
plot(0,0,'-r',0,0,'-b');
if length(NRL) > 0
  plot(NRL(:,[1 2])',repmat([1200 1200],length(NRL),1)','.-r');
end
if length(nts) > 0
  plot(nts(:,[1 2])',repmat([1000 1000],length(nts),1)','.-b');
end
hold off
legend('annotated','found')
title('NT labels');
xlabel('time / s');
ylabel('freq / Hz');
if PAUSE
  pause;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function code = labtype2code(type)
% Map name of label file type to integer code
code = strmatch(type, {'3col','ldc','segmap'}, 'exact') - 1;
if code < 0
  error(['Labeltype ''',type,''' not recognized (not ldc/segmap/3col)']);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [out,outstem] = ensuresuffix(in,ext)
% Append <ext> to <in>, unless it already ends that way.
% outstem returns out less the extension, regardless of when it was added.
lenext = length(ext);
lenin = length(in);
if strcmp(ext, in(max(1,lenin-(lenext-1)):end)) == 0
  out = [in,ext];
else
  out = in;
end
outstem = out(1:end-(lenext-1));
