%GMM_DISTANCE  Compute distance between two Gaussian mixture models.
%
% SYNTAX
%   [distancematrix, options] = gmm_distance(models1, models2, options ...)
%
% DESCRIPTION
%   Computes pairwise distances between a set of Gaussian mixture models.
%
% INPUT
%   models1, models2:
%     Struct arrays of gmmstructs as returned by isp_gmmtrain.
%   options ...:
%     The following parameters can be set as field names in structs, or
%     be specified as field/value pairs:
%     type:
%       A string indicating how to compute the distance. Possibilities:
%       'montecarlokl':
%         Kullback-Leibler distance computed by Monte Carlo sampling.
%       'fastkl':
%         Calculate a fast, but less accurate approximation of the KL-distance.
%       'exactkl':
%         Exact KL-distance. Only possible with a single mixture.
%       'euclidean':
%         The L2 distance between the GMMs.
%       'correlation':
%         Inner product between the GMM's probability density
%         functions normalized to unit L2 norm, i.e., the correlation
%         coefficient or a continuous version of the cosine between
%         vectors. The actual output is one minut the inner product to
%         ensure that smaller is better.
%       Default is 'exactkl' for a single Gaussian and 'fastkl' for mixtures.
%     nSamples:
%       An integer specifying the number of samples when 'montecarlokl' is
%       chosen as type.
%       Default is 200.
%     weightThreshold:
%       When doing Monte Carlo sampling, do not generate random samples from
%       Gaussian mixtures with weight less than this value.
%       Default is 0.01.
%     silent:
%       Boolean specifying if information about the progress shall be
%       suppresed or not. Default: false.
%     forceDiagonal:
%       Boolean specifying if Gaussians with full covariance matrices shall be
%       converted to diagonal matrices. Provided for test purposes.
%       true: Convert full covariance matrices to diagonal.
%       false: Leave covariance matrices untouched. 
%       Default is false.
%     cache1, cache2:
%       Caching of e.g. matrix inverses for 'models1' and 'models2',
%       respectively. In reality, isp_gmmdistance not only return two
%       parameters, but actually return [distancematrix, options, cache1,
%       cache2]. If e.g. models2 never change, one can speed things up
%       by giving the returned cache2 as input. This feture is not
%       heavily tested.
%
% OUTPUT
%   Matrix of distances. Element (i,j) is the distance between the GMMs in
%   models1(i) and models2(j).
%
% EXAMPLE
%     distancematrix = isp_computedistancematrix(models)
%     distancematrix = isp_computedistancematrix([gmm1, gmm2, gmm3], options)
%
% SEE ALSO
%   isp_gmmprobability, isp_gmmrand, isp_gmmtrain.
%
% HISTORY
%   2006-03-31: JHJ - First version for the ISP toolbox.
%   2006-05-07: JHJ - Modified to accept two cell arrays of feature vectors
%   2007-01-22: JHJ - Added 'correlation' distance type
%   2007-04-01: JHJ - Added 'emd' prefix and added optimized code for
%                    'correlation' for diagonal covariance matrices.

% This program is free software; you can redistribute it and/or modify it
% under the terms of the GNU General Public License version 2 as published
% by the Free Software Foundation.



function [distancematrix, options, cache1, cache2]=gmm_distance(models1, models2, varargin)

if nargin < 2
%  error('Not enough input arguments.')
  models2 = models1;
end
    
%    if ~iscell(models1)
%        models1=(models1);
%    end

%    if ~iscell(models2)
%        models2=(models2);
%    end

%    if ~(strcmp(models1(1).type, 'gmm') && strcmp(models2(1).type, 'gmm'))
%        error('Inputs are not Gaussian mixture models.')
%    end

twoSets = ~isequal(models1, models2);

[options.nSamples, options.weightThreshold, options.silent, ...
 options.type, options.forceDiagonal, options.euclideanweight, ...
 options.cache1, options.cache2] = ...
    process_options(varargin, ...
                    'nSamples', 200, ...
                    'weightThreshold', 0.01, ...
                    'silent', true, ...
                    'type', '', ...
                    'forceDiagonal', false, ...
                    'euclideanweight', [], ...
                    'cache1', struct, ...
                    'cache2', struct);
    
nSongs1=length(models1);
nSongs2=length(models2);
nDim=size(models1(1).mean,1);
nMixtures1=size(models1(1).mean,2);
nMixtures2=size(models2(1).mean,2);
mcSamples = options.nSamples;

cache1 = options.cache1;
cache2 = options.cache2;

if isempty(options.type)
  if nMixtures1 > 1 || nMixtures2 > 1
    options.type = 'fastkl';
  else
    options.type = 'exactkl';
  end
end

if options.forceDiagonal
  for n=1:nSongs1
    models1(n).cov = diag(models1(n).cov);
    models1(n).diag = 1;
  end
  
  for n=1:nSongs2
    models2(n).cov = diag(models2(n).cov);
    models2(n).diag = 1;
  end
end
    
type=options.type;
wThreshold = options.weightThreshold;
silent = options.silent;


switch type
 case 'montecarlokl'
  distancematrix=zeros(nSongs2, nSongs1);
  distancematrix2=zeros(nSongs1, nSongs2);
  mfccVector=zeros(nDim, mcSamples*nSongs1);
  probs=zeros(mcSamples*nSongs1,1);
  for iSong=1:nSongs1
    g=models1(iSong);
    g.weight(g.weight<wThreshold )=0;
    mfccVector(:, (iSong-1)*mcSamples+1:iSong*mcSamples) = ...
        gmmrand(g, mcSamples, true);
  end
  
  if twoSets
    % Calculate distance to each song
    selfProb1 = zeros(nSongs1, 1);
    selfProb2 = zeros(nSongs2, 1);
    
    for iSong=1:nSongs1
      selfProb1(iSong) = mean(log( mysp_gmmprobability(models1(iSong), ...
                                                       mfccVector(:, (iSong-1)*mcSamples+1:iSong*mcSamples) ) ));
    end

    for iSong=1:nSongs2
      if ~silent, fprintf('Song %d of %d\n', iSong, nSongs2); end
      probs=log(mysp_gmmprobability(models2(iSong), mfccVector));
      distancematrix2(:,iSong)=mean(reshape(probs, mcSamples, nSongs1),1);
    end
    
    mfccVector=zeros(nDim, mcSamples*nSongs2);
    probs=zeros(mcSamples*nSongs2,1);
    for iSong=1:nSongs2
      g=models2(iSong);
      g.weight(g.weight<wThreshold )=0;
      mfccVector(:, (iSong-1)*mcSamples+1:iSong*mcSamples)=mysp_gmmrand(g, mcSamples, true);
      selfProb2(iSong) = mean(log( mysp_gmmprobability(models2(iSong), ...
                                                       mfccVector(:, (iSong-1)*mcSamples+1:iSong*mcSamples) ) ));
    end
  end

  % Calculate distance to each song
  for iSong=1:nSongs1
    if ~silent, fprintf('Song %d of %d\n', iSong, nSongs1); end
    probs=log(mysp_gmmprobability(models1(iSong), mfccVector));
    distancematrix(:,iSong)=mean(reshape(probs, mcSamples, nSongs2),1);
  end
  
  % Normalize
  if twoSets
    distancematrix = - distancematrix' - distancematrix2 ...
        + repmat(selfProb1,1,nSongs2) ...
        + repmat(selfProb2',nSongs1,1);
  else
    distancematrix = - distancematrix-distancematrix' ...
        + repmat(diag(distancematrix),1,nSongs1) ...
        + repmat(diag(distancematrix)',nSongs1,1);
  end
  
  
  
 case 'fastkl'
  distancematrix=zeros(nSongs2, nSongs1);
  distancematrix2=zeros(nSongs1, nSongs2);
  nSamples=nMixtures1;
  mfccVector=zeros(nDim, nSamples*nSongs1);
  probs=zeros(nSamples*nSongs1,1);
  weightMatrix=zeros(nSongs1,nSamples*nSongs1);
  mlen = 0; %dpwe: modified for unequal numbers of components
            %across models (post prune/merging)
  mlens = zeros(nSongs1+1);
  for iSong=1:nSongs1
    g=models1(iSong);
    m=g.mean;
    w=g.weight;
    len = size(m,2);
    midx = mlen + [1:len];
    mfccVector(:, midx)=m;
    weightMatrix(iSong,midx)=w;
    mlen = mlen + len; % dpwe
    mlens(iSong+1) = mlen;
  end
  mfccVector = mfccVector(:, 1:mlen);
  weightMatrix = weightMatrix(:, 1:mlen);
  
  % Calculate distance to each song
  if twoSets
    selfProb1 = zeros(nSongs1, 1);
    selfProb2 = zeros(nSongs2, 1);
    
    for iSong=1:nSongs1
      selfProb1(iSong) = models1(iSong).weight ...
          * log( mysp_gmmprobability(models1(iSong), ...
                                     mfccVector(:, mlens(iSong)+1:mlens(iSong+1))));
    end
    
    for iSong=1:nSongs2
      if ~silent, fprintf('Song %d of %d\n', iSong, nSongs2); end
      probs=log(mysp_gmmprobability(models2(iSong), mfccVector));
      distancematrix2(:,iSong)=weightMatrix * probs;
    end
    
    nSamples=nMixtures2;
    mfccVector=zeros(nDim, nSamples*nSongs2);
    probs=zeros(nSamples*nSongs2,1);
    weightMatrix=zeros(nSongs2, nSamples*nSongs2);

    mlen = 0; %dpwe: modified for unequal numbers of components
              %across models (post prune/merging)
    
    for iSong=1:nSongs2
      g=models2(iSong);
      m=g.mean;
      w=g.weight;
      midx = mlen + [1:size(m,2)];
      mfccVector(:, midx)=m;
      weightMatrix(iSong,midx)=w;
      selfProb2(iSong) = w*log( mysp_gmmprobability(models2(iSong), ...
                                                    mfccVector(:, midx) ) );
      mlen = mlen + size(m,2); % dpwe
    end
    mfccVector = mfccVector(:, 1:mlen);
    weightMatrix = weightMatrix(:, 1:mlen);
  end
  % Calculate distance to each song
  for iSong=1:nSongs1
    if ~silent, fprintf('Song %d of %d\n', iSong, nSongs1); end
    probs=log(mysp_gmmprobability(models1(iSong), mfccVector));
    %distancematrix(:,iSong)=sum(reshape(probs, nSamples, nSongs2).*weightMatrix,1)';
    distancematrix(:,iSong)=weightMatrix * probs;
  end
  % Normalize
  if twoSets
    distancematrix = - distancematrix' - distancematrix2 ...
        + repmat(selfProb1,1,nSongs2) ...
        + repmat(selfProb2',nSongs1,1);
  else
    distancematrix = - distancematrix-distancematrix' ...
        + repmat(diag(distancematrix),1,nSongs1) ...
        + repmat(diag(distancematrix)',nSongs1,1);
  end
  
  
 case {'correlation', 'euclidean'}
  distancematrix=zeros(nSongs1, nSongs2);
  
  euclidean = strcmp(type, 'euclidean');
  
  if ~twoSets
    models2 = models1;
    nMixtures2 = nMixtures1;
  end
  
  if euclidean
    norm1 = zeros(nSongs1, 1);
    norm2 = zeros(nSongs2, 1);
  end
  
  if models1(1).diag && models2(1).diag
    % Diagonal covariance matrices with 1 mixture
    
    % Normalize to unit norm
    for iSong1 = 1:nSongs1
      weight = models1(iSong1).weight;
      cov = models1(iSong1).cov;
      means = models1(iSong1).mean;
      norm = 0;
      for kMix1 = 1:nMixtures1
        for lMix1 = 1:nMixtures1
          covSum = cov(:,kMix1) + cov(:,lMix1);
          meanDiff = (means(:,kMix1)-means(:,lMix1));
          norm = norm ...
                 +  weight(kMix1) * weight(lMix1) ...
                 * 1/(sqrt(prod(2*pi*covSum ))) ...
                 * exp( -.5*( (meanDiff./covSum).'*meanDiff) );
        end
      end
      if euclidean
        norm1(iSong1) = norm;
      else
        models1(iSong1).weight = weight/sqrt(norm);
      end
    end
    
    % Normalize to unit norm
    for iSong2 = 1:nSongs2
      weight = models2(iSong2).weight;
      cov = models2(iSong2).cov;
      means = models2(iSong2).mean;
      norm = 0;
      for kMix2 = 1:nMixtures2
        for lMix2 = 1:nMixtures2
          covSum = (cov(:,kMix2) + cov(:,lMix2));
          meanDiff = (means(:,kMix2)-means(:,lMix2));
          norm = norm ...
                 +  weight(kMix2) * weight(lMix2) ...
                 * 1/(sqrt(prod(2*pi*covSum ))) ...
                 * exp( -.5*( (meanDiff./covSum).'*meanDiff) );
        end
      end
      if euclidean
        norm2(iSong2) = norm;
      else
        models2(iSong2).weight = weight/sqrt(norm);
      end
    end
    
    if nMixtures1==1 && nMixtures2==1
      temp = cat(1, models1(:));
      cov1 = cat(2, temp.cov);
      mean1 = cat(2, temp.mean);
      weight1 = cat(2, temp.weight);
      
      temp = cat(1, models2(:));
      cov2 = cat(2, temp.cov);
      mean2 = cat(2, temp.mean);
      weight2 = cat(2, temp.weight);
      
      for iSong1 = 1:nSongs1
        covSum = cov1(:,iSong1*ones(nSongs2,1)) + cov2;
        meanDiff = mean1(:,iSong1*ones(nSongs2,1)) - mean2;
        distancematrix(iSong1, :) = ...
            weight1(iSong1) * (weight2 ...
                               ./(sqrt(prod(2*pi*covSum,1))) ...
                               .* exp( -.5*sum((meanDiff.^2)./covSum,1)));
      end
      
    else
      
      
      mix1indices = repmat(1:nMixtures1, 1, nMixtures2);
      mix2indices = reshape(repmat(1:nMixtures1, nMixtures2, 1), 1, ...
                            nMixtures1*nMixtures2);
      for iSong1 = 1:nSongs1
        weight1 = models1(iSong1).weight;
        cov1 = models1(iSong1).cov;
        mean1 = models1(iSong1).mean;
        for jSong2 = 1:nSongs2
          weight2 = models2(jSong2).weight;
          cov2 = models2(jSong2).cov;
          mean2 = models2(jSong2).mean;
          
          covSum = cov1(:, mix1indices) + cov2(:, mix2indices);
          meanDiff = mean1(:, mix1indices) - mean2(:, mix2indices);
          distancematrix(iSong1, jSong2) = sum( ...
              weight1(mix1indices).* weight2(mix2indices) ...
              ./sqrt(prod(2*pi*covSum, 1)) ...
              .* exp( -.5*sum( meanDiff.^2./covSum, 1)) ...
              );
        end
      end
    end
    
  else
    
    % Make sure both feature sets have full covariance matrices
    if models1(1).diag
      covariance = zeros([nDim nDim nMixtures1]);
      for n=1:nSongs1
        for m=1:nMixtures1
          covariance(:,:,m) = diag(models1(n).cov(:, m));
        end
        models1(n).cov = covariance;
      end
    end
    if models2(1).diag
      covariance = zeros([nDim nDim nMixtures2]);
      for n=1:nSongs2
        for m=1:nMixtures2
          covariance(:,:,m) = diag(models2(n).cov(:, m));
        end
        models2(n).cov = covariance;
      end
    end
    
    % Compute norm (correlation: also normalize to unit norm)
    for iSong1 = 1:nSongs1
      weight = models1(iSong1).weight;
      cov = models1(iSong1).cov;
      means = models1(iSong1).mean;
      norm = 0;
      for kMix1 = 1:nMixtures1
        for lMix1 = 1:nMixtures1
          covSum = cov(:,:,kMix1) + cov(:,:,lMix1);
          meanDiff = (means(:,kMix1)-means(:,lMix1));
          norm = norm ...
                 +  weight(kMix1) * weight(lMix1) ...
                 * 1/(sqrt(det(2*pi*covSum ))) ...
                 * exp( -.5*( meanDiff.'*(covSum\meanDiff)) );
        end
      end
      if euclidean
        norm1(iSong1) = norm;
      else
        models1(iSong1).weight = weight/sqrt(norm);
      end
    end
    
    % Compute norm (correlation: also normalize to unit norm)
    for iSong2 = 1:nSongs2
      weight = models2(iSong2).weight;
      cov = models2(iSong2).cov;
      means = models2(iSong2).mean;
      norm = 0;
      for kMix2 = 1:nMixtures2
        for lMix2 = 1:nMixtures2
          covSum = (cov(:,:,kMix2) + cov(:,:,lMix2));
          meanDiff = (means(:,kMix2)-means(:,lMix2));
          norm = norm ...
                 +  weight(kMix2) * weight(lMix2) ...
                 * 1/(sqrt(det(2*pi*covSum ))) ...
                 * exp( -.5*( meanDiff.'*(covSum\meanDiff)) );
        end
      end
      if euclidean
        norm2(iSong2) = norm;
      else
        models2(iSong2).weight = weight/sqrt(norm);
      end
    end
    
    for iSong1 = 1:nSongs1
      weight1 = models1(iSong1).weight;
      cov1 = models1(iSong1).cov;
      mean1 = models1(iSong1).mean;
      for jSong2 = 1:nSongs2
        weight2 = models2(jSong2).weight;
        cov2 = models2(jSong2).cov;
        mean2 = models2(jSong2).mean;
        for kMix1 = 1:nMixtures1
          for lMix2 = 1:nMixtures2
            covSum = cov1(:,:,kMix1) + cov2(:,:,lMix2);
            meanDiff = (mean1(:,kMix1)-mean2(:,lMix2));
            distancematrix(iSong1, jSong2) = ...
                distancematrix(iSong1, jSong2) ...
                +  weight1(kMix1) * weight2(lMix2) ...
                * 1/(sqrt(det(2*pi*covSum ))) ...
                * exp( -.5*( meanDiff.'*(covSum\meanDiff)) );
          end
        end
      end
    end
  end
  
  if euclidean
    % Add the squared norms to the double of the product
    distancematrix = repmat(norm1(:), 1, nSongs2) ...
        + repmat(norm2(:)', nSongs1, 1) - 2*distancematrix;
  else
    % Create distance measure where 0 is equality and 1 is uncorrelated
    distancematrix = 1 - distancematrix;
  end
  
  
  
  
 case 'exactkl'
  if nMixtures1~=1 || ( twoSets && nMixtures2~=1 )
    error(['Unable to compute exact KL-distance for other ' ...
           'numbers of mixtures than 1.'])
  end
  
  if ~twoSets
    models2 = models1;
  end
  
  % Things could be done much more efficiently for diagonal covariance
  % matrices, but for now, they are converted to full matrices.
  
  if models1(1).diag && models2(1).diag
    
    % Precompute stuff
    mean1 = zeros([nDim nSongs1]);
    
    for iSong=1:nSongs1
      mean1(:,iSong) = models1(iSong).mean;
    end
    
    % Compute non-symmetric KL-distance if two sets of feature
    % vectors were passed as arguments
    if twoSets
      mean2 = zeros([nDim nSongs2]);
      for iSong=1:nSongs2
        mean2(:,iSong) = models2(iSong).mean;
      end
      
      invCov1 = zeros([nDim nSongs1]);
      for iSong=1:nSongs1
        invCov1(:,iSong) = 1./models1(iSong).cov;
      end
      
      distancematrix2 = zeros(nSongs1, nSongs2);
      intermediateResult1 = sum(mean1.^2.*invCov1, 1).';
      intermediateResult2 = mean1.*invCov1;
      
      for iSong=1:nSongs2
        igmmCov = models2(iSong).cov;
        distancematrix2(:, iSong) = ...
            invCov1.'*igmmCov + ...
            invCov1.'*(mean2(:, iSong).^2) + ...
            intermediateResult1 - ...
            intermediateResult2.'*(2*mean2(:, iSong)) - ...
            nDim;
      end
      distancematrix2 = .5*distancematrix2;
    else
      mean2=mean1;
    end
    
    % Compute non-symmetric KL-distance
    invCov2 = zeros([nDim nSongs2]);
    for iSong=1:nSongs2
      invCov2(:,iSong) = 1./models2(iSong).cov;
    end
    
    distancematrix = zeros(nSongs1, nSongs2);
    intermediateResult1 = sum(mean2.^2.*invCov2, 1).';
    intermediateResult2 = mean2.*invCov2;
    
    for iSong=1:nSongs1
      if ~silent, fprintf('Song %d of %d\n', iSong, nSongs1); end
      igmmCov = models1(iSong).cov;
      
      distancematrix(:, iSong) = ...
          invCov2.'*igmmCov + ...
          invCov2.'*(mean1(:, iSong).^2) + ...
          intermediateResult1 - ...
          intermediateResult2.'*(2*mean1(:, iSong)) - ...
          nDim;
      
    end
    distancematrix = .5*distancematrix;
  else
    % Full covariance matrix
    
    % Handle the very special case when only one of the feature sets have full
    % covariance matrix.
    if models1(1).diag
      for n=1:nSongs1
        models1(n).cov = diag(models1(n).cov);
      end
    end
    
    if models2(1).diag
      for n=1:nSongs2
        models2(n).cov = diag(models2(n).cov);
      end
    end
    
    % Precompute stuff
    if isfield(cache1, 'mean1')
      mean1 = cache1.mean1;
    else
      mean1 = zeros([nDim nSongs1]);
      for iSong1=1:nSongs1
        mean1(:,iSong1) = models1(iSong1).mean;
      end
    end
    
    if isfield(cache2, 'invCov2')
      invCov2 = cache2.invCov2;
    else
      invCov2 = zeros([nDim nDim nSongs2]);
      for iSong2=1:nSongs2
        invCov2(:,:,iSong2) = inv(models2(iSong2).cov);
      end
    end
    
    cache1.mean1 = mean1;
    cache2.invCov2 = invCov2;
    
    if isfield(cache2, 'mean2')
      mean2 = cache2.mean2;
    else
      mean2 = zeros([nDim nSongs2]);
      for iSong2=1:nSongs2
        mean2(:,iSong2) = models2(iSong2).mean;
      end
      cache2.mean2 = mean2;
    end
    
    if twoSets
      if isfield(cache1, 'invCov1')
        invCov1 = cache1.invCov1;
      else
        invCov1 = zeros([nDim nDim nSongs1]);
        for iSong1=1:nSongs1
          invCov1(:,:,iSong1) = ...
              inv(models1(iSong1).cov);
        end
      end
      cache1.invCov1 = invCov1;
      
      % Compute non-symmetric KL-distance if two sets of feature
      % vectors were passed as arguments
      intermediateResult1 = zeros(nSongs1,1);
      intermediateResult2 = zeros([nSongs1 nDim]);
      for iSong=1:nSongs1
        intermediateResult2(iSong,:) = mean1(:,iSong).'*invCov1(:,:,iSong);
        intermediateResult1(iSong) = intermediateResult2(iSong,:) ...
            * mean1(:,iSong);
      end
      
      distancematrix2 = zeros(nSongs1, nSongs2);
      for iSong=1:nSongs2
        if ~silent, fprintf('Song %d of %d\n', iSong, nSongs2); end
        m2sqr = mean2(:, iSong)*mean2(:, iSong)';
        distancematrix2(:, iSong) = ...
            reshape(sum(sum(invCov1.*models2(iSong).cov(:,:,ones(nSongs1,1)))), [], 1) + ...
            reshape(sum(sum(invCov1.*m2sqr(:,:,ones(nSongs1,1)))), [], 1) + ...
            intermediateResult1 - ...
            intermediateResult2*(2*mean2(:, iSong)) - ...
            nDim;
      end
      distancematrix2 = .5*distancematrix2;
    end
    
    % Compute non-symmetric KL-distance
    intermediateResult1 = zeros(nSongs2,1);
    intermediateResult2 = zeros([nSongs2 nDim]);
    
    for iSong=1:nSongs2
      intermediateResult2(iSong,:) = mean2(:,iSong).'*invCov2(:,:,iSong);
      intermediateResult1(iSong) = intermediateResult2(iSong,:) ...
          * mean2(:,iSong);
    end
    
    distancematrix = zeros(nSongs2, nSongs1);
    
    for iSong=1:nSongs1
      if ~silent, fprintf('Song %d of %d\n', iSong, nSongs1); end
      m1sqr = mean1(:, iSong)*mean1(:, iSong)';
      distancematrix(:, iSong) = ...
          reshape(sum(sum(invCov2.*models1(iSong).cov(:,:,ones(nSongs2,1)))), [], 1) + ...
          reshape(sum(sum(invCov2.*m1sqr(:,:,ones(nSongs2,1)))), [], 1) + ...
          intermediateResult1 - ...
          intermediateResult2*(2*mean1(:, iSong)) - ...
          nDim;
      
    end
    
    distancematrix = .5*distancematrix;
  end
  
  % Add non-symmetric distances to get the symmetric distance
  if twoSets
    distancematrix = distancematrix.' + distancematrix2;
  else
    distancematrix = distancematrix.' + distancematrix;
  end
  
  
  
 otherwise
  error(['Invalid type-argument '''  type '''.'])
end



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%ISP_GMMRAND  Generate random vectors from a Gaussian Mixture Model.
%
% SYNTAX
%   vectors=isp_gmmrand(gmmstruct, nVectors, enforceMix)
%
% DESCRIPTION
%   Generate random vectors from a Gaussian mixture model.
%
% INPUT
%   gmmstruct:
%     Struct with the following fields (as returned by isp_gmmtrain):
%     weight:
%       Vector of probabilities of each Gaussian mixture.
%     mean:
%       Matrix of column vectors containing the means of the individual
%       multivariate Gaussian components.
%     diagonal:
%       True if covariance matrices are diagonal. False otherwise.
%     covariance:
%       If 'diagonal' is true, a 2-dimensional matrix where covariance(:,k)
%       is the diagonal of the k'th covariance matrix. Otherwise it is a
%       3-dimensional matrix where covariance(:,:,k) is the covariance
%       matrix of the k'th multivariate Gaussian.
%   nVectors:
%     The number of random vectors to return (default: 1).
%   enforceMix:
%     If false (default), let the choice of Gaussian be random,
%     distributed according to 'weight'. If true, let the number of
%     vectors from each mixture be proportional to 'weight'.
%
% OUTPUT
%   vectors:
%     Matrix where the columns consist of random vectors.
%
% SEE ALSO
%   isp_gmmprobability, isp_gmmtrain.
%
% HISTORY
%   Created by Jesper H. Jensen.

% This program is free software; you can redistribute it and/or modify it
% under the terms of the GNU General Public License version 2 as published
% by the Free Software Foundation.


function vectors=mysp_gmmrand(gmmstruct, nVectors, enforceMix)
    weights=gmmstruct.weight;
    means=gmmstruct.mean;
    covariances=gmmstruct.cov;
    diagonal=gmmstruct.diag;

    %
    % Interpret arguments
    %
    
    if ~exist('nVectors', 'var')
        nVectors=1;
    end

    if ~exist('enforceMix', 'var')
        enforceMix=false;
    end

    if size(weights,2) > size(weights,1)
        weights=weights';
    end

    
    %
    % Allocate memory
    %
    
    vecDim=size(means,1);
    vectors=zeros(vecDim,nVectors);

    
    if enforceMix
        thresholds=(0:nVectors)/nVectors;
    else
        thresholds=rand(1,nVectors);
    end
    cumulativeWeights=[0; cumsum(weights)];
    
    
    if diagonal
        covarianceSqrt=sqrt(covariances);
    else
        covarianceSqrt=repmat(nan,size(covariances));
    end
    
    for n=1:nVectors

        % Use a uniformly distributed random number to determine which of the
        % Gaussian mixtures to use
        thisThreshold=thresholds(n);

        % Do a binary search
        maxIndex=length(cumulativeWeights);
        minIndex=1;
        while maxIndex-minIndex ~= 1
            middleIndex=ceil(0.5*(maxIndex + minIndex));
            if cumulativeWeights(middleIndex) < thisThreshold
                minIndex=middleIndex;
            else
                maxIndex=middleIndex;
            end
        end
        mixtureNumber=maxIndex-1;

        if diagonal
            vectors(:,n)=covarianceSqrt(:,mixtureNumber).*randn(vecDim,1) ...
                + means(:,mixtureNumber);
        else
            % Only calculate the "square root" of a correlation matrix once.
            if isnan(covarianceSqrt(1,1,mixtureNumber))
                covarianceSqrt(:,:,mixtureNumber) = ...
                    sqrtm(covariances(:,:,mixtureNumber));
            end

            % Calculate a vector from a multivariate Gaussian distribution by
            % multiplying the "square root" of the correlation matrix by a vector of
            % independent Gaussians and add the mean.
            vectors(:,n)=covarianceSqrt(:,:,mixtureNumber)*randn(vecDim,1) ...
                + means(:,mixtureNumber);
        end
    end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%ISP_GMMPROBABILITY  Probability of vectors according to a specified Gaussian Mixture Model.
%
% SYNTAX
%   prob=isp_gmmprobability(gmmstruct, vectors)
%
% DESCRIPTION
%   Return the probability of vectors assuming the specified Gaussian
%   mixture model.
%
% INPUT
%   gmmstruct:
%     Struct with the following fields:
%     weight:
%       Vector of probabilities of each Gaussian mixture.
%     mean:
%       Matrix of column vectors containing the means of the individual
%       multivariate Gaussian components.
%     covariance:
%       3-dimensional matrix which contain the correlation matrices of
%       the individual multivariate Gaussian components.
%   vectors:
%     Cell array containing column vectors or matrix of column vectors.
%
% OUTPUT
%   prob:
%     Vector of probabilities of the vectors specified in 'vectors'
%
% SEE ALSO
%   isp_gmmrand, isp_gmmtrain.
%
% HISTORY
%   Created by Jesper H. Jensen 2005

% This program is free software; you can redistribute it and/or modify it
% under the terms of the GNU General Public License version 2 as published
% by the Free Software Foundation.

function prob=mysp_gmmprobability(gmmstruct, vectors)
    
    % The number of mean value vectors to subtract at a time.
    blocksize=500;

    weight=gmmstruct.weight;
    meanVecs=gmmstruct.mean;
    diagonal=gmmstruct.diag;
    
    nVectors=size(vectors,2);
    vecDim=size(vectors,1);
    nMixtures=size(meanVecs,2);
    
    if size(weight,2) > size(weight,1)
        weight=weight';
    end

    if iscell(vectors)
        vectors=cell2mat(vectors);
    end

    
    if diagonal
        invDiagonals=1./gmmstruct.cov;
        detCovariance=prod(gmmstruct.cov, 1);
    else
        detCovariance=zeros(nMixtures,1);
        inverseCovariance=zeros(size(gmmstruct.cov));
        for n=1:nMixtures
            detCovariance(n)=det(gmmstruct.cov(:,:,n));
            inverseCovariance(:,:,n)=inv(gmmstruct.cov(:,:,n));
        end
    end

    % Allocate variables
    prob=zeros(nVectors,1);
    exponents=zeros(1, nVectors);
    zeroMeanVectors=zeros(size(vectors));

    if diagonal
        vectorsSquared = vectors.^2;
    end

    % Calculate probabilities one mixture at a time
    for n=1:nMixtures
        % Ignore this mixture if the covariance matrix is singular
        if detCovariance(n)==0
            warning('Ignoring singular covariance matrix.')
            continue
        end
        constantFactor = 1./sqrt( ((2*pi)^vecDim)*detCovariance(n));

        if diagonal
            % Diagonal covariance
            d = -.5*invDiagonals(:,n);
            exponents = d'*vectorsSquared + d'*meanVecs(:,n).^2 - 2*(d.*meanVecs(:,n))'*vectors;
            prob = prob + (weight(n)*constantFactor)*exp(exponents');
        else
            % Full covariance
            if nVectors <= blocksize
                zeroMeanVectors = vectors - meanVecs(:,n*ones(nVectors,1));
            else
                tempVec=meanVecs(:,n*ones(blocksize,1));
                for m=1:blocksize:nVectors-blocksize;
                    zeroMeanVectors(:,m:m+blocksize-1)=vectors(:,m:m+blocksize-1)-tempVec;
                end
                zeroMeanVectors (:,m+blocksize:end)=vectors(:,m+blocksize:end)-meanVecs(:,n*ones(nVectors-m-blocksize+1,1));
            end
        
            exponents = -.5*sum(zeroMeanVectors.*(inverseCovariance(:,:,n)*zeroMeanVectors),1);
            prob = prob + (weight(n)*constantFactor)*exp(exponents');
        end
    end
