function model = gmm_mergecomponents(modelin, volthresh, klthresh, VERBOSE)
% modelout = gmm_mergecomponents(modelin, volthresh, klthresh)
%    For tight clusters (determinant within volthresh of minimum)
%    merge them with components whose kl distance is no more than klthresh.
%    New version uses KL distance between clusters as basis.
% 20120-05-17, 2012-03-13 Dan Ellis dpwe@ee.columbia.edu

if nargin < 2; volthresh = 2.0; end
if nargin < 3; klthresh = 20; end
if nargin < 4; VERBOSE = 0; end

model = modelin;

[ndim,ncomp] = size(model.mean);

% display "volume" of gaussians
for i = 1:ncomp
  vol(i) = det(squeeze(model.cov(:,:,i))).^(1/ndim);
end

% merge all the low-variance clusters
[v,smallest] = min(vol);

%smallests = find(vol < volthresh*v);

% just keep the small ones AND process them in order of volume
[vvs,smallests] = sort(vol,'ascend');
smallests = smallests(vvs < volthresh*v);

for i = 1:length(smallests);

  smix = smallests(i);
  if smix > 0
    if VERBOSE; disp(['Merging model ', num2str(smix)]); end
    [model,mmap] = merge_cluster(model, smix, klthresh, VERBOSE);
    % map smallests list into new models, preserving zeros
    mmap = [0 mmap];
    smallests = mmap(1+smallests);
  end

end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [modelout,mmap] = merge_cluster(model,tomerge, KLTHRESH, VERBOSE)
% [modelout,mmap] = merge_cluster(model,tomerge, KLTHRESH, VERBOSE)
%    Take a set of Gaussian models model.{mean(i,:),cov(:,:,i),weight(i)}
%    and merge together all models within threshold of
%    model(tomerge)
%    Return reduced model space, and a map of how old model indices
%    map to new model (0 = deleted).
% 2012-02-28 Dan Ellis dpwe@ee.columbia.edu

% Clusters whose Gaussian likelihood under the seed cluster (with its
% covariance doubled) exceeds this are merged in to it
%MINGPROB = 0.1;

[ndim,ncomp] = size(model.mean);

% Evalulate all other centers under gauss of tomerge
%gp = gaussian_prob(model.mean, model.mean(:,tomerge), ...
%                   2*squeeze(model.cov(:,:,tomerge))).^(1/ndim);

gexp = gmm_explode(model);
gd = gmm_distance(gexp(tomerge), gexp);

if VERBOSE;   gmm_print(model, gd, tomerge);  end

modeix = find(gd < KLTHRESH);

if VERBOSE; disp(['Mode centers: ',sprintf('%d ',modeix)]); end

% Merge members of all those clusters to the tomerge mode
if length(modeix) > 1
  model = gmmcomp_merge(model, modeix);
  tomerge = modeix(1); % bug 2012-05-21
end

% Eliminate other models
keepix = ones(1,ncomp);
keepix(modeix) = 0;  
keepix(tomerge) = 1;
mmap = cumsum(keepix);
mmap(find(keepix==0)) = 0; % map deleted elements to zero
unkeepix = find(~keepix);
model = gmmcomp_remove(model,unkeepix);
tomergenew = mmap(tomerge);
newone = min(find(mmap > 0));

% make tomerge index 1
model = gmmcomp_swap(model, tomergenew, 1);

mmap([newone tomerge]) = mmap([tomerge newone]);

modelout = model;
% if VERBOSE;   gmm_print(modelout); end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function modelout = gmmcomp_merge(model, comps)
% modelout = merge_gaussians(model, comps)
%    model gaussian mixture model structure with i mixtures; 
%    model.mean(:,i), model.cov(:,:,i), and model.weight(i).
%    modelout is returned as a similar Gaussian descriptor that
%    would result from merging the gaussian components indexed by comps.
% 2012-02-29 Dan Ellis dpwe@ee.columbia.edu

% Map members of all those clusters to the tomerge mode
Sx = zeros(size(model.mean(:,1)));
Sxy = zeros(size(model.cov(:,:,1)));
N = 0;

for i = 1:length(comps)
  % Combine model
  ci = comps(i);
  wi = model.weight(ci);
  mi = model.mean(:,ci);
  Mx = mi * wi;
  Sx = Sx + Mx;
  Sxy = Sxy + wi*(squeeze(model.cov(:,:,ci)) + mi*mi');
  N = N + wi;
end
modelout = model;
mo = Sx/N;
modelout.mean(:,comps(1)) = mo;
modelout.cov(:,:,comps(1)) = Sxy/N - mo*mo';
modelout.weight(comps(1)) = N;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function modelout = gmmcomp_remove(model, comps)
ncomp = size(model.mean,2);
keepix = ones(1,ncomp);
keepix(comps) = 0;
keepix = find(keepix);
modelout = model;
modelout.mean = modelout.mean(:,keepix);
modelout.cov = modelout.cov(:,:,keepix);
modelout.weight = modelout.weight(keepix);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function modelout = gmmcomp_swap(model, comp1, comp2)
modelout = model;
modelout.mean(:,[comp1 comp2]) = modelout.mean(:,[comp2 comp1]);
modelout.cov(:,:,[comp1 comp2]) = modelout.cov(:,:,[comp2 comp1]);
modelout.weight([comp1 comp2]) = modelout.weight([comp2 comp1]);
