function [L,IX,XTRA] = read_label_file(LABFILE,IXIN,FMT,DFLTLAB,VERBOSE)
% [L,IX,XTRA] = read_label_file(F,IXIN,FMT,DFLT,VERB)
%    Read SAD labels from label file
%    L is 3-column rows: start_time end_time label_index
%    IX returns the cell array of actual labels that L(:,3) indexes
%    into.
%    Option IXIN is cell array of labels to predefine ordering.
%    FMT=0 for 3-column (default), 1 for LDC 11 column, 2 for 4 col segmap
%    XTRA returns a full set of 11 columns for each row (for LDC
%    files)
%    DFLT, if specified, is a label to assign to empty labels ('').
%    VERBOSE causes verbose output if set.
% 2012-02-29 Dan Ellis dpwe@ee.columbia.edu

if nargin < 2;  IXIN = []; end
if nargin < 3;  FMT = 0; end
if nargin < 3;  DFLTLAB = ''; end
if nargin < 5;  VERBOSE = 0; end

LDCLABELS = 0;
SEGMAP = 0;

if FMT == 1
  LDCLABELS = 1;
elseif FMT == 2
  SEGMAP = 1;
end

if exist(LABFILE,'file') == 0
  error(['read_label_file: file ',LABFILE,' not found']);
end

[p,cleanstem,e] = fileparts(LABFILE);

XTRA = cell(0,1);

% Read in the input label file
if LDCLABELS || SEGMAP
  if LDCLABELS
    ddisp(sprintf('Reading LDC-format label file %s ...', LABFILE), ...
          VERBOSE);
      % LDC2011E32 tab-separated columns
      % 1:filename 2:channel 3:starttime 4:endtime 5:timesource
      % 6:spkr 7:spkrsource 8:lang 9:langsource 10:transcript 11:transsource
      [fn,ch,ss,ee,ll,sid,ssrc,lid,lsrc,tsrc,lasrc] = textread(LABFILE, ...
                             '%s\t%s\t%f\t%f\t%s\t%s\t%s\t%s\t%s\t%s\t%s', ...
                             'delimiter', '\t');
      if length(ll) == 0
        error(['Trying to read ',LABFILE, ...
               ' as LDC (11 column) format, but not enough columns']);
      end
      for i = 1:length(fn)
        XTRA{i} = {fn{i},ch{i},ss(i),ee(i),ll{i},sid{i},ssrc{i},lid{i},lsrc{i},tsrc{i},lasrc{i}};
        % debug - save original times in other fields
        %        XTRA{i} = {fn{i},ch{i},ss(i),ee(i),ll{i},sid{i},ssrc{i},lid{i},num2str(ss(i)),num2str(ee(i)),ll{i}};
      end
  else % SEGMAP
    ddisp(sprintf('Reading 4 column segmap file %s ...', LABFILE), VERBOSE);
    % compositefilename starttime endtime componentfilename
    [fn,ss,ee,ll] = textread(LABFILE,'%s\t%f\t%f\t%s','delimiter','\t');
  end
  % match from end of filenames
  %goodlines = strmatch(cleanstem,fn);
  nf = cellfun(@fliplr,fn,'UniformOutput',false);
  goodlines = strmatch(fliplr(cleanstem),nf);
  ss = ss(goodlines);
  ee = ee(goodlines);
  ll = ll(goodlines);
else
  ddisp(sprintf('Reading 3 column label file %s ...', LABFILE), VERBOSE);
  [ss,ee,ll] = textread(LABFILE,'%f\t%f\t%s','delimiter','\t');
end

ddisp(['got ', num2str(length(ss)), ' labels'], VERBOSE);

if length(IXIN) == 0
  IX = unique(ll);
else
  IX = IXIN;
end

lix = zeros(0,1);
for i = 1:length(ll);
  lab = ll{i};
  if length(lab) == 0; lab = DFLTLAB; end
  lixitem = strmatch(lab,IX,'exact');
  if length(lixitem)==0
    error(['label "',lab,'" (line ',num2str(i),') not in IXIN']);
  end
  lix(i) = lixitem;
end

L = [ss,ee,lix'];


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function ddisp(msg, verb)
if verb > 0
  disp(msg);
end
% verb also carries file descriptor of log file
if verb ~= 0
  if verb ~= 1 && verb ~= 2
    fprintf(abs(verb), '%s\n', msg);
  end
end
